from collections import OrderedDict
from typing import Any, Dict, Iterator, List, Tuple, Union

from antlr4.tree.Tree import TerminalNode, Token, Tree
from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid

from _qpd_antlr import QPDParser as qp
from qpd._parser.sql import QPDSql
from triad.utils.schema import unquote_name
from qpd.constants import AGGREGATION_FUNCTIONS, JOIN_TYPES
from qpd.dataframe import DataFrame, DataFrames
from qpd.workflow import QPDWorkflow


class ColumnMention(object):
    def __init__(self, df_name: str, col_name: str):
        self.df_name = df_name
        self.col_name = col_name
        self.encoded = "_" + to_uuid(df_name, col_name).split("-")[-1]
        self.expr = df_name + "." + col_name

    def __repr__(self) -> str:
        return self.expr

    def __hash__(self) -> int:
        return hash((self.df_name, self.col_name))

    def __eq__(self, other: Any) -> bool:
        return self.df_name == other.df_name and self.col_name == other.col_name


class ColumnMentions(object):
    def __init__(self, *data: Any):
        self._data: Dict[ColumnMention, int] = OrderedDict()
        self.add(*data)

    def add(self, *data: Any) -> "ColumnMentions":
        for x in data:
            if isinstance(x, ColumnMention):
                self._data[x] = 1
            elif isinstance(x, ColumnMentions):
                self._data.update(x._data)
            elif isinstance(x, tuple):
                self._data[ColumnMention(x[0], x[1])] = 1
            else:
                raise ValueError(f"{x} is invalid")
        return self

    def __contains__(self, m: ColumnMention) -> bool:
        return m in self._data

    def __iter__(self) -> Iterator[ColumnMention]:
        return iter(self._data.keys())

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return ",".join(x.__repr__() for x in self)

    @property
    def single(self) -> ColumnMention:
        assert_or_throw(len(self._data) == 1, f"multiple mentions: {self._data}")
        return next(iter(self._data.keys()))


class Join(object):
    def __init__(
        self,
        df_name: str,
        df: DataFrame,
        join_type: str = "",
        natural: bool = False,
        criteria_context: Any = None,
    ):
        self.df_name = df_name
        self.df = df
        self.join_type = join_type
        self.natural = natural
        self.criteria_context = criteria_context
        self.conditions: List[Any] = []

    def set_conditions(self, cond: List[Any]):
        self.conditions = cond


class VisitorContext(object):
    def __init__(self, *args: Any, **kwargs: Any):
        self._data: Dict[str, Any] = {}
        for v in args:
            if isinstance(v, VisitorContext):
                # self._data.update(v._data)
                self._data = v._data
            elif isinstance(v, Dict):
                self._data.update(v)
            else:
                raise ValueError(f"{v} is invalid")
        self._data.update(kwargs)
        if "node_to_token" not in self._data:
            self._data["node_to_token"] = {}

    def set(self, key: str, value: Any) -> None:
        self._data[key] = value

    def get(self, key: str, default: Any = None) -> Any:
        return self._data.get(key, default)

    @property
    def workflow(self) -> QPDWorkflow:
        return self._data["workflow"]

    @property
    def sql(self) -> QPDSql:
        return self._data["sql"]

    @property
    def current(self) -> DataFrame:
        return self._data["current"]

    def update_current(self, df: DataFrame) -> DataFrame:
        self._data["current"] = df
        return df

    @property
    def dfs(self) -> DataFrames:
        return self._data["dfs"]

    def add_column_mentions(self, ctx: Any) -> ColumnMentions:
        if "col_mentions" in self._data and id(ctx) in self._data["col_mentions"]:
            return self.get_column_mentions(ctx)
        expr = self.to_str(ctx, "", unquote=False)
        assert_or_throw(expr != "*", ValueError("'*' shouldn't be used here"))
        mentions = ColumnMentions(*self.dfs._resolver.get_cols(expr))
        if "col_mentions" not in self._data:
            self._data["col_mentions"] = {id(ctx): mentions}
        else:
            self._data["col_mentions"][id(ctx)] = mentions
        if "all_mentions" not in self._data:
            self._data["all_mentions"] = ColumnMentions()
        self._data["all_mentions"].add(mentions)
        return mentions

    def get_column_mentions(self, ctx: Any) -> ColumnMentions:
        return self._data["col_mentions"][id(ctx)]  # type:ignore

    @property
    def all_column_mentions(self) -> ColumnMentions:
        v = self._data.get("all_mentions", None)
        if v is None:
            return ColumnMentions()
        return v

    def add_join_relation(self, join: Join) -> None:
        if "joins" not in self._data:
            self._data["joins"] = []
        for j in self._data["joins"]:
            assert_or_throw(
                j.df_name != join.df_name,
                ValueError(f"{join.df_name} is already in join line"),
            )
        self._data["joins"].append(join)

    @property
    def joins(self) -> List[Join]:
        return self.get("joins", [])

    def set_col_encoded(self) -> None:
        return self.set("col_encoded", True)

    @property
    def col_encoded(self) -> bool:
        return self.get("col_encoded", False)

    def ctx_to_col_names(self, ctx: Any) -> List[str]:
        expr = self.to_str(ctx, "")
        cols = self.dfs.name_resolver.get_cols(expr)
        if self.col_encoded:
            return [self._obj_to_col_name(x) for x in cols]
        else:
            return [x[0] + "." + x[1] for x in cols]

    def to_str(
        self,
        node: Union[Tree, Token, str, None],
        delimit: str = " ",
        unquote: bool = False,
    ) -> str:
        if isinstance(node, str):
            return node
        if node is None:
            return ""
        res = delimit.join(
            self._get_node_str(node, self._data["node_to_token"], unquote=unquote)[1]
        )
        if unquote:
            return res
        return res.rstrip(" ")

    def expr(self, node: Union[Tree, Token, None]) -> str:
        return self.to_str(node, " ")

    def _obj_to_col_name(self, obj: Any):
        return "_" + to_uuid(obj).split("-")[-1]

    def _get_node_str(
        self,
        node: Union[Tree, Token],
        cache: Dict[int, Tuple[Token, List[str]]],
        unquote: bool = False,
    ) -> Tuple[Token, List[str]]:
        if unquote and isinstance(node, qp.QuotedIdentifierContext):
            _, ls = self._get_node_str(node, cache, unquote=False)
            ss = "".join(ls)
            return node, [unquote_name(ss)]
        if id(node) in cache:
            return cache[id(node)]
        elif isinstance(node, Token):
            res = (node, [self.sql.raw_code[node.start : node.stop + 1]])
        elif isinstance(node, TerminalNode):
            res = self._get_node_str(node.getSymbol(), cache, unquote=unquote)
        else:
            strs: List[str] = []
            first: Any = None
            for i in range(node.getChildCount()):
                t, arr = self._get_node_str(node.getChild(i), cache, unquote=unquote)
                if first is None:
                    first = t
                strs += arr
            res = (first, strs)
        cache[id(node)] = res
        return res


def normalize_join_type(join_type: str) -> str:
    orig = join_type
    join_type = "_".join(x for x in join_type.strip().lower().split(" ") if x != "")
    if join_type == "" or join_type == "_":
        return "inner"
    if join_type in ["left", "right", "full"]:
        return join_type + "_outer"
    if join_type in ["semi", "anti"]:
        return "left_" + join_type
    assert_or_throw(join_type in JOIN_TYPES, ValueError(f"{orig} is invalid"))
    return join_type


def is_agg(func: str) -> bool:
    return func.lower() in AGGREGATION_FUNCTIONS



from qpd_test.tests_base import TestsBase
from typing import Any, Tuple, Iterable
from pandasql import sqldf
from timeit import timeit


class BenchmarkTests(object):
    class Tests(TestsBase):
        def _test_select(
            self, *sizes: int, repeat: int = 1
        ) -> Iterable[Tuple[int, float, float]]:
            for n in sizes:
                a = self.make_rand_df(
                    n,
                    a=int,
                    b=str,
                    c=float,
                    d=int,
                    e=bool,
                    f=str,
                    g=str,
                    h=float,
                    i=str,
                    j=int,
                    k=bool,
                    l=float,
                )
                q, s = self._time_vs_sqlite(
                    repeat,
                    """
                    SELECT l AS ll,k AS kk,j AS jj,h AS hh,f,e,b,(a+l)*(a-l) AS a FROM a
                    """,
                    a=a,
                )
                yield n, q, s

        def _test_select_where(
            self, *sizes: int, repeat: int = 1
        ) -> Iterable[Tuple[int, float, float]]:
            for n in sizes:
                a = self.make_rand_df(
                    n,
                    a=int,
                    b=str,
                    c=float,
                    d=int,
                    e=bool,
                    f=str,
                    g=str,
                    h=float,
                    i=str,
                    j=int,
                    k=bool,
                    l=float,
                )
                q, s = self._time_vs_sqlite(
                    repeat,
                    """
                    SELECT l AS ll,k AS kk,j AS jj,h AS hh,f,e,b,(a+l)*(a-l) AS a FROM a
                    WHERE l<10 OR e OR (c<0.5 AND h>0.5)
                    """,
                    a=a,
                )
                yield n, q, s

        def _test_integration_1(
            self, *sizes: int, repeat: int = 1
        ) -> Iterable[Tuple[int, float, float]]:
            for n in sizes:
                a = self.make_rand_df(
                    n, a=int, b=str, c=float, d=int, e=bool, f=str, g=str, h=float
                )
                q, s = self._time_vs_sqlite(
                    repeat,
                    """
                    WITH
                        a1 AS (
                            SELECT a+1 AS a, b, c FROM a
                        ),
                        a2 AS (
                            SELECT a,MAX(b) AS b_max, AVG(c) AS c_avg FROM a GROUP BY a
                        ),
                        a3 AS (
                            SELECT d+2 AS d, f, g, h FROM a WHERE e
                        )
                    SELECT a1.a,b,c,b_max,c_avg,f,g,h FROM a1
                        INNER JOIN a2 ON a1.a=a2.a
                        LEFT JOIN a3 ON a1.a=a3.d
                    """,
                    a=a,
                )
                yield n, q, s

        def _time_vs_sqlite(self, n: int, sql: str, **dfs: Any) -> Tuple[float, float]:
            data: Any = self.to_dfs(dfs)
            qpd_time = timeit(lambda: self._run(sql, data), number=n) / n
            data = self.to_pd_dfs(dfs)
            sqlite_time = timeit(lambda: sqldf(sql, data), number=n) / n
            return qpd_time, sqlite_time



import pandas as pd
from pytest import raises

from qpd.dataframe import DataFrames
from qpd._parser.sql import QPDSql
from qpd._parser.utils import (ColumnMention, ColumnMentions, Join,
                               VisitorContext, is_agg, normalize_join_type)
from qpd_pandas import QPDPandasEngine
from qpd.workflow import QPDWorkflow, QPDWorkflowContext


def test_column_mention():
    a = ColumnMention("a", "b")
    b = ColumnMention("a", "c")
    assert not a == b
    assert a == ColumnMention("a", "b")
    assert "a" == a.df_name
    assert "b" == a.col_name
    assert "_d5f618372100" == a.encoded
    assert "a.b" == a.__repr__()
    s = {a, b}
    assert ColumnMention("a", "c") in s
    assert ColumnMention("a", "d") not in s


def test_column_mentions():
    a = ColumnMention("a", "b")
    m1 = ColumnMentions(a, ("a", "c"), a)
    m2 = ColumnMentions(m1, m1, ColumnMention("b", "a"))
    assert 3 == len(m2)
    assert "a.b,a.c,b.a" == m2.__repr__()
    raises(AssertionError, lambda: m2.single)
    m3 = ColumnMentions(a)
    assert m3.single is a
    raises(ValueError, lambda: ColumnMentions("a.b"))


def test_visitor_context():
    raises(ValueError, lambda: VisitorContext(1))
    v = VisitorContext()
    v.set("x", 1)
    v2 = VisitorContext(v, z=3)
    assert v._data is v2._data
    assert 1 == v2.get("x", 2)
    assert 2 == v2.get("y", 2)
    assert 3 == v2.get("z", 2)
    df1 = _make_df([[0, 1]], ["a", "b"])
    df2 = _make_df([[0, 2]], ["a", "c"])
    dfs = DataFrames(a=df1, b=df2)
    v = VisitorContext(v2, current=df1, dfs=dfs)
    assert v.current is df1
    assert v.dfs is dfs
    sql = QPDSql("SELECT * \n FROM  a", "singleStatement")
    wf = QPDWorkflow(QPDWorkflowContext(QPDPandasEngine(), dfs))
    v = VisitorContext(v, dict(sql=sql, workflow=wf))
    assert v.sql is sql
    assert v.workflow is wf
    v.update_current(df2)
    assert v.current is df2

    assert 0 == len(v.all_column_mentions)
    obj = "a.*"
    m = v.add_column_mentions(obj)
    assert ColumnMention("a", "a") in m
    assert ColumnMention("a", "b") in m
    assert len(m) == 2
    assert v.get_column_mentions(obj) is m
    obj = "c"
    m = v.add_column_mentions(obj)
    m = v.add_column_mentions(obj)  # will read from cache
    assert ColumnMention("b", "c") in m
    assert len(m) == 1
    assert 3 == len(v.all_column_mentions)

    assert not v.col_encoded
    v.set_col_encoded()
    assert v.col_encoded

    assert "" == v.to_str(None)
    assert "SELECT * FROM a" == v.expr(sql.tree)
    assert "SELECT * FROM a" == v.expr(sql.tree)  # test caching

    sql = QPDSql("a.*", "expression")
    v = VisitorContext(v, sql=sql)
    v.set("col_encoded", False)
    assert ["a.a", "a.b"] == v.ctx_to_col_names(sql.tree)
    v.set_col_encoded()
    assert ['_845ca596baa3', '_d5f618372100'] == v.ctx_to_col_names(sql.tree)

    assert 0 == len(v.joins)
    v.add_join_relation(Join("a", df1))
    v.add_join_relation(Join("b", df2))
    raises(ValueError, lambda: v.add_join_relation(Join("a", df2)))
    assert 2 == len(v.joins)
    v.joins[1].set_conditions([[None, None]])


def test_is_agg():
    assert is_agg("COUNT")
    assert not is_agg("SIN")


def test_normalize_join_type():
    assert "inner" == normalize_join_type(" ")
    assert "inner" == normalize_join_type("")
    assert "inner" == normalize_join_type(" INNER ")
    assert "left_outer" == normalize_join_type("LEFT")
    assert "full_outer" == normalize_join_type("full")
    assert "left_semi" == normalize_join_type("semi")
    assert "left_anti" == normalize_join_type("left   anti")
    raises(ValueError, lambda: normalize_join_type("xyz"))


def _make_df(arr, cols):
    return QPDPandasEngine().to_df(pd.DataFrame(arr, columns=cols))



from qpd._parser.sql import QPDSql
from pytest import raises


def test_qpd_sql():
    sql = QPDSql("SELECT * FROM a WHERE x=1", "singleStatement")
    assert "SELECT * FROM a WHERE x=1" == sql.raw_code
    assert "SELECT * FROM a WHERE x=1" == sql.code
    with raises(SyntaxError):
        QPDSql("select * from a where x=1", "singleStatement", ignore_case=False)
    sql = QPDSql("select * from a where x=1", "singleStatement", ignore_case=True)
    assert "select * from a where x=1" == sql.raw_code
    assert "SELECT * FROM a WHERE x=1" == sql.code



from typing import Any
import ray
ray.init(ignore_reinit_error=True)
import os
os.environ["MODIN_ENGINE"] = "ray"
import modin.pandas as pd
import pandas

from qpd.dataframe import Column, DataFrame
from qpd_ray import QPDRayEngine
from qpd_test.engine_suite import QPDEngineTests
from qpd_test import assert_df_eq
from qpd.specs import AggFunctionSpec


class EngineTests(QPDEngineTests.Tests):
    def test_sort(self):
        return

    def test_agg_first_last(self):
        def assert_eq(df, gp_keys, gp_map, expected, expected_cols):
            res = self.to_pandas_df(self.op.group_agg(df, gp_keys, gp_map))
            assert_df_eq(res, expected, expected_cols)

        pdf = self.to_df([[0, 1], [0, 1], [1, None], [1, 5]], ["a", "b"])
        assert_eq(
            pdf,
            ["a"],
            {
                "b1": ("b", AggFunctionSpec("first", unique=False, dropna=False)),
                "b2": ("b", AggFunctionSpec("first", unique=False, dropna=True)),
                "b3": ("b", AggFunctionSpec("last", unique=False, dropna=False)),
                "b4": ("b", AggFunctionSpec("last", unique=False, dropna=True)),
            },
            [[1, 1, 1, 1], [5, 5, 5, 5]],
            ["b1", "b2", "b3", "b4"],
        )

    def test_window_row_number(self):
        return

    def test_window_ranks(self):
        return

    def test_window_lead_lag(self):
        return

    def test_window_sum(self):
        return

    def to_pandas_df(self, df: DataFrame) -> pandas.DataFrame:
        if isinstance(df, DataFrame):
            return self.op.to_native(df)._to_pandas()
        elif isinstance(df, pd.DataFrame):
            return df._to_pandas()
        elif isinstance(df, pandas.DataFrame):
            return df
        else:
            raise ValueError(df)

    def to_pandas_series(self, col: Column) -> pandas.Series:
        return pandas.Series(col.native._to_pandas(), name=col.name)

    def make_qpd_engine(self):
        #ray.init(ignore_reinit_error=True)
        return QPDRayEngine()

    def to_native_df(self, data: Any, columns: Any) -> Any:
        if isinstance(data, pd.DataFrame):
            return data
        if isinstance(data, pandas.DataFrame):
            return pd.DataFrame(data)
        return pd.DataFrame(data, columns=columns)



import pandas as pd
from pytest import raises
from qpd._parser.sql import QPDSql
from qpd._parser.utils import VisitorContext
from qpd._parser.visitors import PreScan, StatementVisitor
from qpd.workflow import QPDWorkflow, QPDWorkflowContext, WorkflowDataFrame
from qpd_pandas import QPDPandasEngine
from qpd_test import assert_df_eq


def test_pre_scan():
    def assert_eq(dfs, sql, mentions):
        sql = QPDSql(sql, "singleStatement")
        ctx = QPDWorkflowContext(QPDPandasEngine(), dfs)
        wf = QPDWorkflow(ctx)
        v = PreScan(
            VisitorContext(),
            sql=sql,
            workflow=wf,
            current=wf.dfs[list(dfs.keys())[0]],
            dfs=wf.dfs,
        )
        v.visit(sql.tree)
        m = set(x.__repr__() for x in v.all_column_mentions)
        if mentions == "":
            assert len(m) == 0
        else:
            assert set(mentions.split(",")) == m
        return v

    def not_support(dfs, sql):
        with raises(NotImplementedError):
            assert_eq(dfs, sql, "")

    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    c = pd.DataFrame([[0, 1, 2, 3, 4, 5]], columns=["a", "b", "c", "d", "e", "f"])
    v = assert_eq(dict(a=a, b=b), "SELECT a.* FROM a", "a.a,a.b")
    assert set(v.get("dfs").keys()) == {"a"}
    assert not v.get("has_agg_func", False)
    v = assert_eq(dict(a=a, b=b), "SELECT * FROM a", "")
    assert v.get("select_all", False)
    assert_eq(dict(c=c), "SELECT a+1+b AS x,b AS y FROM c", "c.a,c.b")
    v = assert_eq(
        dict(a=a, b=b, c=c),
        """
        SELECT a.a AS x,SUM(c.b) AS y
        FROM c INNER JOIN a
            ON c.b=a.b
        WHERE e=10
        GROUP BY a.b
        HAVING SUM(c.c)>0
        ORDER BY c.d
        LIMIT 10
        """,
        "a.a,a.b,c.b,c.c,c.d,c.e",
    )
    assert set(v.get("dfs").keys()) == {"a", "c"}
    assert v.get("has_agg_func", False)
    v = assert_eq(
        dict(a=a, b=b, c=c),
        """
        SELECT a, LEAD(e,1) OVER(PARTITION BY b ORDER BY c) FROM c
        """,
        "c.a,c.b,c.c,c.e",
    )
    assert v.get("has_window", False)
    v = assert_eq(
        dict(c=c),
        "SELECT a+1+b AS x,SUM(b) AS y FROM c AS cc WHERE e=10",
        "cc.a,cc.b,cc.e",
    )
    assert set(v.get("dfs").keys()) == {"cc"}
    v = assert_eq(
        dict(c=c),
        "SELECT a+1+b AS x,SUM(b) AS y FROM (c AS cc) AS x WHERE e=10",
        "x.a,x.b,x.e",
    )
    assert set(v.get("dfs").keys()) == {"x"}

    # prescan argument types
    v = assert_eq(
        dict(c=c),
        """SELECT f(a, a+a, a+1, -1 , 1+1) AS x,
                  g('str', NOT a, sin( sin(a))) AS x FROM c""",
        "c.a",
    )
    d = {v.to_str(x[0], ""): (x[1], x[2]) for x in v._func_args}
    assert {
        "a": (True, True),
        "a+a": (True, False),
        "a+1": (True, False),
        "-1": (False, True),
        "1+1": (False, False),
        "'str'": (False, True),
        "NOTa": (True, False),
        "sin(a)": (True, False),
        "sin(sin(a))": (True, False),
    } == d

    not_support(dict(c=c), "SELECT *, SUM(a) FROM c")
    not_support(
        dict(c=c), "SELECT SUM(a), ROW_NUMBER() OVER(PARTITION BY b ORDER BY c) FROM c"
    )

    # special chars
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a b", "-"])
    assert_eq(dict(a=a), "SELECT a.* FROM a", "a.a b,a.-")
    assert_eq(dict(a=a), "SELECT `a b`, `-` FROM a", "a.a b,a.-")

    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["", " "])
    assert_eq(dict(a=a), "SELECT a.* FROM a", "a.,a. ")
    assert_eq(dict(a=a), "SELECT ` `, `` FROM a", "a. ,a.")


def test_pre_scan_join():
    def assert_eq(dfs, sql, *joins):
        sql = QPDSql(sql, "singleStatement")
        ctx = QPDWorkflowContext(QPDPandasEngine(), dfs)
        wf = QPDWorkflow(ctx)
        v = PreScan(
            VisitorContext(),
            sql=sql,
            workflow=wf,
            current=wf.dfs[list(dfs.keys())[0]],
            dfs=wf.dfs,
        )
        v.visit(sql.tree)
        for e, a in zip(joins, v.joins[1:]):
            c = [[v.to_str(x[0], ""), v.to_str(x[1], "")] for x in a.conditions]
            assert e == [a.df_name, a.join_type, a.natural] + c
        return v

    def bad(dfs, sql):
        with raises(Exception):
            assert_eq(dfs, sql)

    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    c = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "d"])
    d = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "e"])
    e = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "f"])
    dfs = dict(a=a, b=b, c=c, d=d, e=e)
    assert_eq(
        dfs,
        """
        SELECT * FROM a
        """,
    )
    assert_eq(
        dfs,
        """
        SELECT a.*,c FROM a JOIN b ON a.a=b.a
        """,
        ["b", "inner", False, ["a.a", "b.a"]],
    )
    # order of on comparison will be corrected
    assert_eq(
        dfs,
        """
        SELECT a.*,c FROM a LEFT JOIN b ON c=a.a
        """,
        ["b", "left_outer", False, ["a.a", "c"]],
    )
    # alias of table can be handled
    assert_eq(
        dfs,
        """
        SELECT aa.*,c FROM a AS aa FULL JOIN b AS bb ON bb.c=aa.a
        """,
        ["bb", "full_outer", False, ["aa.a", "bb.c"]],
    )
    # multiple joins and multiple conditions
    assert_eq(
        dfs,
        """
        SELECT a.*,b FROM a
            JOIN b ON b.a=a.a
            SEMI JOIN c ON a.a=c.a
            ANTI JOIN d ON e = b AND d.a=a.a
        """,
        ["b", "inner", False, ["a.a", "b.a"]],
        ["c", "left_semi", False, ["a.a", "c.a"]],
        ["d", "left_anti", False, ["b", "e"], ["a.a", "d.a"]],
    )
    # multiple tables in condition
    assert_eq(
        dfs,
        """
        SELECT a.*,b FROM a
            JOIN b ON b.a=a.a
            RIGHT JOIN c ON c.a=a.a+b.c
        """,
        ["b", "inner", False, ["a.a", "b.a"]],
        ["c", "right_outer", False, ["a.a+b.c", "c.a"]],
    )
    # cross join
    assert_eq(
        dfs,
        """
        SELECT a.*,c FROM a CROSS JOIN b
        """,
        ["b", "cross", False],
    )
    # natural join
    assert_eq(
        dfs,
        """
        SELECT a.*,c FROM a NATURAL LEFT JOIN b
        """,
        ["b", "left_outer", True],
    )

    # # * is ambiguous. this should be captured after prescan
    # bad(dfs, """
    # SELECT * FROM a JOIN c ON a.a=c.a
    # """)
    # wrong condition
    bad(
        dfs,
        """
    SELECT a.* FROM a JOIN c ON a.a=b.a
    """,
    )
    # wrong condition
    bad(
        dfs,
        """
    SELECT a.* FROM a JOIN c ON a.a=a.a
    """,
    )
    # invalid condition
    bad(
        dfs,
        """
    SELECT a.* FROM a JOIN c ON a.a>b.a
    """,
    )


def test_op_precedence():
    a = pd.DataFrame([[1, 4], [2, 5]], columns=["a", "b"])
    assert_eq(
        dict(a=a),
        "SELECT a+1 AS a, b FROM a",
        [[2, 4], [3, 5]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a),
        "SELECT a+2*3 AS a, b FROM a",
        [[7, 4], [8, 5]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a),
        "SELECT a+1-2*3/4 AS a, b FROM a",
        [[0.5, 4], [1.5, 5]],
        ["a", "b"],
    )


def test_const_select():
    assert_eq(
        {},
        "SELECT 1 AS a, 1.5 AS b, TRUE AS c",
        [[1, 1.5, True]],
        ["a", "b", "c"],
    )
    assert_eq(
        {},
        "SELECT 1+2 AS a, 1.5*3 AS b, 'x' AS c",
        [[3, 4.5, "x"]],
        ["a", "b", "c"],
    )


def test_case_when():
    a = pd.DataFrame([[1, "a", 4], [2, "a", 5], [2, "b", 6]], columns=["a", "b", "c"])
    assert_eq(
        dict(a=a),
        """
        SELECT a,b,
            CASE
                WHEN a=1 THEN 10+c
                WHEN b='b' AND a=2 THEN 20+c
                ELSE 30*2
            END AS c
        FROM a
        """,
        [[1, "a", 14], [2, "a", 60], [2, "b", 26]],
        ["a", "b", "c"],
    )


def test_is_value():
    a = pd.DataFrame([[True], [False], [None]], columns=["a"])
    assert_eq(
        dict(a=a),
        """
        SELECT a,
            NOT a AS b,
            a IS NULL AS c,
            a IS NOT NULL AS d,
            a IS TRUE AS e,
            a IS NOT TRUE AS f,
            a IS FALSE AS g,
            a IS NOT FALSE AS h,
            CASE
                WHEN a IS NULL THEN 0
                WHEN a IS TRUE THEN 1
                ELSE 2
            END AS i
        FROM a
        """,
        [
            [True, False, False, True, True, False, False, True, 1],
            [False, True, False, True, False, True, True, False, 2],
            [None, None, True, False, False, True, False, True, 0],
        ],
        list("abcdefghi"),
    )


def test_is_in():
    a = pd.DataFrame([[0, 1, 1], [2, 3, 4]], columns=["a", "b", "c"])
    assert_eq(
        dict(a=a),
        """
        SELECT *,
            c IN (1,5) AS d,
            c NOT IN (1,5) AS e,
            c IN (7+2,a+b) AS f
        FROM a
        """,
        [[0, 1, 1, True, False, True], [2, 3, 4, False, True, False]],
        list("abcdef"),
    )


def test_is_between():
    a = pd.DataFrame([[0, 1, 1], [2, 3, 4], [1, 2, None]], columns=["a", "b", "c"])
    assert_eq(
        dict(a=a),
        """
        SELECT *,
            c BETWEEN 0 AND 1 AS d,
            c NOT BETWEEN 0 AND 1 AS e,
            c BETWEEN a AND b AS f,
            c NOT BETWEEN a AND b AS g
        FROM a
        """,
        [
            [0, 1, 1, True, False, True, False],
            [2, 3, 4, False, True, False, True],
            [1, 2, None, None, None, None, None],
        ],
        list("abcdefg"),
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM a
        WHERE c BETWEEN 0 AND 1
        """,
        [[0, 1, 1]],
        list("abc"),
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM a
        WHERE c NOT BETWEEN 0 AND 1
        """,
        [[2, 3, 4]],
        list("abc"),
    )


def test_order_by_limit():
    a = pd.DataFrame([[1, "a"], [2, "b"], [None, "a"]], columns=["a", "b"])
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM a ORDER BY a LIMIT 1*1
        """,
        [[None, "a"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM a ORDER BY a NULLS LAST LIMIT 1*1
        """,
        [[1, "a"]],
        ["a", "b"],
    )


def test_drop_duplicates():
    a = pd.DataFrame([[1, "a"], [2, "b"], [1, "a"]], columns=["a", "b"])
    b = pd.DataFrame([[1, "b"], [4, "bb"]], columns=["a", "b"])
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT ALL * FROM a
        """,
        [[1, "a"], [2, "b"], [1, "a"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT DISTINCT * FROM a
        """,
        [[1, "a"], [2, "b"]],
        ["a", "b"],
    )


def test_from():
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    c = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "d"])
    d = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "e"])
    e = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "f"])
    dfs = dict(a=a, b=b, c=c, d=d, e=e)
    assert_eq(
        dfs,
        """
        SELECT * FROM a
        """,
        [[0, 1], [0, 2], [1, 3]],
        ["a", "b"],
    )
    assert_eq(
        dfs,
        """
        SELECT x.* FROM a AS x
        """,
        [[0, 1], [0, 2], [1, 3]],
        ["a", "b"],
    )
    assert_eq(
        dfs,
        """
        SELECT a.*,c FROM a INNER JOIN b ON b.a == a.a
        """,
        [[0, 1, 20], [0, 2, 20]],
        ["a", "b", "c"],
    )
    assert_eq(
        dfs,
        """
        SELECT x.*,c FROM a AS x INNER JOIN b AS y ON y.a == x.a
        """,
        [[0, 1, 20], [0, 2, 20]],
        ["a", "b", "c"],
    )


def test_join():
    # basic operations
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    assert_eq(
        dict(a=a, b=b),
        """SELECT a.*,c AS x FROM a INNER JOIN b ON a.a=b.a""",
        [[0, 1, 20], [0, 2, 20]],
        ["a", "b", "x"],
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT b,c,a.a AS x FROM a INNER JOIN b ON c=b""",
        [[3, 3, 1]],
        ["b", "c", "x"],
    )

    # semi join
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    assert_eq(
        dict(a=a, b=b),
        """SELECT a.* FROM a LEFT SEMI JOIN b ON a.a=b.a""",
        [[0, 1], [0, 2]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT * FROM a LEFT SEMI JOIN b ON a.a=b.a""",
        [[0, 1], [0, 2]],
        ["a", "b"],
    )

    # multiple conditions, and formula equations
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20, 1], [2, 2, 1]], columns=["a", "c", "d"])
    assert_eq(
        dict(a=a, b=b),
        """SELECT b*c AS y,a.a AS x FROM a INNER JOIN b ON c+d=b AND a.a*2=b.a""",
        [[6, 1]],
        ["y", "x"],
    )

    # multiple joins
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[2, 20], [3, 30]], columns=["b", "c"])
    c = pd.DataFrame([[20, 200]], columns=["c", "d"])
    assert_eq(
        dict(a=a, b=b, c=c),
        """SELECT a.*,b.c,c.d FROM a
                INNER JOIN b ON a.b=b.b
                INNER JOIN c ON c.c=b.c""",
        [[0, 2, 20, 200]],
        ["a", "b", "c", "d"],
    )
    assert_eq(
        dict(a=a, b=b, c=c),
        """SELECT a.*,b.c,c.d FROM a
                INNER JOIN b ON a.b=b.b
                INNER JOIN c ON c.c=b.c+(a.a+0)""",
        [[0, 2, 20, 200]],
        ["a", "b", "c", "d"],
    )


def test_where():
    # basic operations
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    assert_eq(
        dict(a=a, b=b), """SELECT * FROM a WHERE b>=2""", [[0, 2], [1, 3]], ["a", "b"]
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT * FROM a WHERE b>=2 AND NOT (a==0)""",
        [[1, 3]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b), """SELECT * FROM a WHERE b>2 AND a!=0""", [[1, 3]], ["a", "b"]
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT a.*,c,TRUE AS t, FALSE AS f FROM a INNER JOIN b ON a.a=b.a
            WHERE a.a<=0 AND a.b*2<3 AND (TRUE OR FALSE)""",
        [[0, 1, 20, True, False]],
        ["a", "b", "c", "t", "f"],
    )

    a = pd.DataFrame([[None, "x", 0.5], [10, None, 0.6]], columns=["a", "b", "c"])
    assert_eq(
        dict(a=a, b=b),
        """SELECT * FROM a WHERE c IS NOT NULL AND (a<50 AND b IS NOT NULL)""",
        [],
        ["a", "b", "c"],
    )


def test_agg():
    # basic
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    assert_eq(
        dict(a=a, b=b),
        """SELECT a, SUM(b) AS b FROM a GROUP BY a""",
        [[0, 3], [1, 3]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT a, SUM(b) AS sb, AVG(b) AS ab, MIN(b) AS mb, MAX(b) AS xb
            FROM a GROUP BY a""",
        [[0, 3, 1.5, 1, 2], [1, 3, 3, 3, 3]],
        ["a", "sb", "ab", "mb", "xb"],
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT a, FIRST(b) AS fb, LAST(b) AS lb
            FROM a GROUP BY a""",
        [[0, 1, 2], [1, 3, 3]],
        ["a", "fb", "lb"],
    )

    # select tests
    a = pd.DataFrame([[0, 1], [0, 2], [3, 2]], columns=["a", "b"])
    assert_eq(
        dict(a=a),
        """SELECT a,SUM(b) AS sa FROM a GROUP BY a
        """,
        [[0, 3], [3, 2]],
        ["a", "sa"],
    )
    assert_eq(
        dict(a=a),
        """SELECT a AS x,SUM(b) AS sa FROM a GROUP BY a
        """,
        [[0, 3], [3, 2]],
        ["x", "sa"],
    )
    a = pd.DataFrame([[0, 1, 4], [0, 2, 5], [3, 2, 6]], columns=["a", "b", "c"])
    assert_eq(
        dict(a=a),
        """SELECT a,SUM(b)+SUM(c) AS s, SUM(b*c) AS t FROM a GROUP BY a
        """,
        [[0, 12, 14], [3, 8, 12]],
        ["a", "s", "t"],
    )
    assert_eq(
        dict(a=a),
        """SELECT a,SUM(b)+a AS s FROM a GROUP BY a
        """,
        [[0, 3], [3, 5]],
        ["a", "s"],
    )

    # where
    assert_eq(
        dict(a=a),
        """SELECT a,SUM(b*b) AS sa FROM a WHERE b=2 GROUP BY a
        """,
        [[0, 4], [3, 4]],
        ["a", "sa"],
    )

    # having
    assert_eq(
        dict(a=a),
        """SELECT a,SUM(b*b) AS sa FROM a WHERE b=2 GROUP BY a
            HAVING SUM(b*b)=4 AND a=0
        """,
        [[0, 4]],
        ["a", "sa"],
    )
    assert_eq(
        dict(a=a),
        """SELECT a,SUM(b*b) AS sa FROM a GROUP BY a
            HAVING SUM(a+b)=3
        """,
        [[0, 5]],
        ["a", "sa"],
    )


def test_agg_count():
    a = pd.DataFrame([[0, 1], [0, 2], [1, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 3]], columns=["a", "c"])
    assert_eq(
        dict(a=a, b=b),
        """SELECT a, COUNT() AS b FROM a GROUP BY a""",
        [[0, 2], [1, 1]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """SELECT a, COUNT(DISTINCT b) AS b FROM a GROUP BY a""",
        [[0, 2], [1, 1]],
        ["a", "b"],
    )
    a = pd.DataFrame(
        [[0, 1, None], [0, 1, None], [0, 4, 1], [1, None, None], [None, None, None]],
        columns=["a", "b", "c"],
    )
    nan = float("nan")
    assert_eq(
        dict(a=a),
        """
        SELECT
            a AS a1,
            COUNT(DISTINCT *) AS a2,
            c AS a3,
            COUNT(*) AS a4,
            MIN(c) AS a5,
            COUNT(DISTINCT b,c) AS a6
        FROM a GROUP BY a
        """,
        [[0, 2, nan, 3, 1, 2], [1, 1, nan, 1, nan, 1], [nan, 1, nan, 1, nan, 1]],
        ["a1", "a2", "a3", "a4", "a5", "a6"],
    )


def test_agg_no_groupby():
    # TODO: should also test this with group by
    a = pd.DataFrame(
        [[0, 1, None], [0, 1, None], [0, 4, 1], [1, None, None], [None, None, None]],
        columns=["a", "b", "c"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT
            FIRST(c) AS a,                  -- None
            FIRST(c IGNORE NULLS) AS b,     -- 1
            LAST(b) AS c,                   -- None
            LAST(b IGNORE NULLS) AS d,      -- 4
            MIN(a) AS e,                    -- 0
            MAX(a) AS f,                    -- 1
            AVG(b) AS g,                    -- 2
            SUM(b) AS h,                    -- 6
            COUNT(b) AS i,                  -- 3
            COUNT(DISTINCT b) AS j,         -- 2
            COUNT(*) AS k,                  -- 5
            COUNT(DISTINCT *) AS l,         -- 4
            COUNT(b, c) AS m,               -- 5
            COUNT(DISTINCT b, c) AS n       -- 3
        FROM a
        """,
        [[None, 1, None, 4, 0, 1, 2, 6, 3, 2, 5, 4, 5, 3]],
        list("abcdefghijklmn"),
    )


def test_join_agg():
    a = pd.DataFrame([[0, 1], [0, 2], [2, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 20], [2, 30]], columns=["a", "c"])
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT a.a,SUM(a.b)+SUM(c) AS x FROM a
        INNER JOIN b ON a.a=b.a
        WHERE c>=20
        GROUP BY a.a
        """,
        [[0, 43], [2, 33]],
        ["a", "x"],
    )
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT a.a AS y,SUM(a.b)+SUM(c) AS x FROM a
        INNER JOIN b ON a.a=b.a
        WHERE c>=20
        GROUP BY a.a
        HAVING SUM(a.b)+SUM(c)<40
        """,
        [[2, 33]],
        ["y", "x"],
    )


def test_window():
    a = pd.DataFrame([[0, 1], [0, 4], [2, 3]], columns=["a", "b"])
    assert_eq(
        dict(a=a),
        """
        SELECT *, ROW_NUMBER() OVER () AS x
        FROM a
        """,
        [[0, 1, 1], [0, 4, 2], [2, 3, 3]],
        ["a", "b", "x"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT *, ROW_NUMBER() OVER (PARTITION BY a) AS x
        FROM a
        """,
        [[0, 1, 1], [0, 4, 2], [2, 3, 1]],
        ["a", "b", "x"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT *, ROW_NUMBER() OVER (ORDER BY b DESC) AS x
        FROM a
        """,
        [[0, 1, 3], [0, 4, 1], [2, 3, 2]],
        ["a", "b", "x"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT *, ROW_NUMBER() OVER (ORDER BY a ASC, b DESC) AS x
        FROM a
        """,
        [[0, 1, 2], [0, 4, 1], [2, 3, 3]],
        ["a", "b", "x"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT *, ROW_NUMBER() OVER (PARTITION BY a ORDER BY a,b DESC) AS x
        FROM a
        """,
        [[0, 1, 2], [0, 4, 1], [2, 3, 1]],
        ["a", "b", "x"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT *,
            ROW_NUMBER() OVER () AS c,
            ROW_NUMBER() OVER (PARTITION BY a) AS d,
            ROW_NUMBER() OVER (ORDER BY b DESC) AS e,
            ROW_NUMBER() OVER (ORDER BY a ASC, b DESC) AS f,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY a,b DESC) AS g
        FROM a
        """,
        [[0, 1, 1, 1, 3, 2, 2], [0, 4, 2, 2, 1, 1, 1], [2, 3, 3, 1, 2, 3, 1]],
        list("abcdefg"),
    )

    # if ordering by single column, the behavior should follow the SQL behavior on nulls
    # if ordering by multiple columns, the behavior may not follow
    # the SQL behavior on nulls because pandas does not support different na_positions
    # for each order by column
    a = pd.DataFrame([[0, 1], [0, 2], [0, None], [None, 3]], columns=["a", "b"])
    assert_eq(
        dict(a=a),
        """
        SELECT *,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY b DESC NULLS FIRST) AS c,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY b DESC NULLS LAST) AS d,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY b NULLS FIRST) AS e,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY b NULLS LAST) AS f,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY b DESC) AS g,
            ROW_NUMBER() OVER (PARTITION BY a ORDER BY b) AS h
        FROM a
        """,
        [
            [0, 1, 3, 2, 2, 1, 2, 2],
            [0, 2, 2, 1, 3, 2, 1, 3],
            [0, None, 1, 3, 1, 3, 3, 1],
            [None, 3, 1, 1, 1, 1, 1, 1],
        ],
        list("abcdefgh"),
    )


def test_inline_table():
    a = pd.DataFrame([[0, 1], [0, 4], [2, 3]], columns=["a", "b"])
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM (SELECT * FROM a) AS b
        """,
        [[0, 1], [0, 4], [2, 3]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM (SELECT a AS aa, b AS bb FROM a) AS b
        """,
        [[0, 1], [0, 4], [2, 3]],
        ["aa", "bb"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM (SELECT a AS aa, b AS bb FROM a) AS a
        """,
        [[0, 1], [0, 4], [2, 3]],
        ["aa", "bb"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM (SELECT aa AS aaa, bb AS bbb FROM
            (SELECT a AS aa, b AS bb FROM a) AS a) AS a
        """,
        [[0, 1], [0, 4], [2, 3]],
        ["aaa", "bbb"],
    )

    assert_eq(
        dict(a=a),
        """
        SELECT * FROM (SELECT a AS aa, b AS bb FROM a)
        """,
        [[0, 1], [0, 4], [2, 3]],
        ["aa", "bb"],
    )
    assert_eq(
        dict(a=a),
        """
        SELECT * FROM (SELECT aa AS aaa, bb AS bbb FROM
            (SELECT a AS aa, b AS bb FROM a))
        """,
        [[0, 1], [0, 4], [2, 3]],
        ["aaa", "bbb"],
    )

    a = pd.DataFrame([[0, 1], [0, 4], [2, 3]], columns=["a", "b"])
    b = pd.DataFrame([[0, 10], [2, 30]], columns=["c", "d"])
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT aa.*,d FROM
            (SELECT * FROM a WHERE a>0) AS aa INNER JOIN
            (SELECT c AS a, d FROM b) AS bb
            ON aa.a=bb.a
        """,
        [[2, 3, 30]],
        ["a", "b", "d"],
    )


def test_union():
    a = pd.DataFrame([["x", "a"], ["x", "a"], [None, None]], columns=["a", "b"])
    b = pd.DataFrame([["xx", "aa"], [None, None], ["a", "x"]], columns=["b", "a"])
    c = pd.DataFrame([["c", "d"]], columns=["a", "b"])
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a UNION SELECT * FROM b
        """,
        [["x", "a"], [None, None], ["xx", "aa"], ["a", "x"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a UNION DISTINCT SELECT * FROM b
        """,
        [["x", "a"], [None, None], ["xx", "aa"], ["a", "x"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a UNION ALL SELECT * FROM b
        """,
        [["x", "a"], ["x", "a"], [None, None], ["xx", "aa"], [None, None], ["a", "x"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b, c=c),
        """
        SELECT * FROM a UNION SELECT * FROM b UNION SELECT * FROM c
        """,
        [["x", "a"], [None, None], ["xx", "aa"], ["a", "x"], ["c", "d"]],
        ["a", "b"],
    )


def test_intersect():
    a = pd.DataFrame([["x", "a"], ["x", "a"], [None, None]], columns=["a", "b"])
    b = pd.DataFrame(
        [["xx", "aa"], [None, None], [None, None], ["a", "x"]], columns=["b", "a"]
    )
    c = pd.DataFrame([["c", "d"], [None, None]], columns=["a", "b"])
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a INTERSECT SELECT * FROM b
        """,
        [[None, None]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a INTERSECT DISTINCT SELECT * FROM b
        """,
        [[None, None]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b, c=c),
        """
        SELECT * FROM a
            INTERSECT ALL SELECT * FROM b
            INTERSECT ALL SELECT * FROM c
        """,
        [[None, None]],
        ["a", "b"],
    )


def test_except():
    a = pd.DataFrame([["x", "a"], ["x", "a"], [None, None]], columns=["a", "b"])
    b = pd.DataFrame(
        [["xx", "aa"], [None, None], [None, None], ["a", "x"]], columns=["b", "a"]
    )
    c = pd.DataFrame([["c", "d"], [None, None]], columns=["a", "b"])
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a EXCEPT SELECT * FROM b
        """,
        [["x", "a"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b),
        """
        SELECT * FROM a EXCEPT DISTINCT SELECT * FROM b
        """,
        [["x", "a"]],
        ["a", "b"],
    )
    assert_eq(
        dict(a=a, b=b, c=c),
        """
        SELECT * FROM a EXCEPT ALL SELECT * FROM b EXCEPT ALL SELECT * FROM c
        """,
        [["x", "a"], ["x", "a"]],
        ["a", "b"],
    )


def test_with():
    a = pd.DataFrame([[1, "a"], [2, "b"]], columns=["a", "b"])
    b = pd.DataFrame([[1, "b"], [4, "bb"]], columns=["a", "b"])
    assert_eq(
        dict(a=a, b=b),
        """
        WITH
            a AS (
                SELECT a AS aa, b AS bb FROM a
            ),
            c AS (
                SELECT aa-1 AS aa, bb FROM a
            )
        SELECT * FROM c UNION SELECT * FROM b
        """,
        [[0, "a"], [1, "b"], [4, "bb"]],
        ["aa", "bb"],
    )


def assert_eq(dfs, sql, expected, expected_cols):
    df = _run(sql, **dfs)
    assert_df_eq(df, expected, expected_cols)


def _run(sql, **kwargs):
    return _debug_run(sql, "singleStatement", StatementVisitor, **kwargs)[0]


def _debug_run(sql, rule, visitor, **kwargs):
    sql = QPDSql(sql, rule)
    ctx = QPDWorkflowContext(QPDPandasEngine(), kwargs)
    wf = QPDWorkflow(ctx)
    v = visitor(VisitorContext(sql=sql, workflow=wf, dfs=wf.dfs))
    wf.assemble_output(v.visit(sql.tree))
    wf.run()
    return ctx.result, v



from collections import OrderedDict, defaultdict
from typing import Any, Dict, Iterable, List, Set, Union, Tuple

from triad.collections import IndexedOrderedDict
from triad.utils.assertion import assert_or_throw
from triad.utils.hash import to_uuid
from triad.utils.schema import unquote_name


class Column(object):
    def __init__(self, data: Any, col: str = ""):
        self._col = col
        self._data = data

    def __uuid__(self) -> str:
        return to_uuid(self.name, id(self.native))

    @property
    def native(self) -> Any:
        return self._data

    @property
    def name(self) -> str:
        return self._col

    def rename(self, name) -> "Column":
        return Column(self.native, name)

    @property
    def has_name(self) -> bool:
        return True

    def assert_has_name(self) -> None:
        assert_or_throw(self.has_name, "column does not have a name")


class DataFrame(IndexedOrderedDict[str, Column]):
    def __init__(self, *args: Any):
        super().__init__()

        def add(objs):
            for a in objs:
                if isinstance(a, Column):
                    a.assert_has_name()
                    assert_or_throw(
                        a.name not in self, ValueError(f"{a.name} already exists")
                    )
                    self[a.name] = a
                elif isinstance(a, DataFrame):
                    for v in a.values():
                        assert_or_throw(
                            v.name not in self, ValueError(f"{v.name} already exists")
                        )
                        self[v.name] = v
                elif isinstance(a, Iterable):
                    add(a)
                else:
                    raise ValueError(f"{a} is not valid")

        add(args)
        self.set_readonly()

    def __getitem__(self, key: Union[str, List[str]]) -> Union[Column, "DataFrame"]:
        if isinstance(key, str):
            if key != "*":
                return super().__getitem__(key)
            return self
        res: List[Column] = []
        for col in key:
            if col == "*":
                res += list(self.values())
            else:
                res.append(super().__getitem__(col))
        return DataFrame(res)


class ColumnNameResolver(object):
    def __init__(self, *args: Any, **kwargs: Any):
        self._data: Dict[str, List[str]] = OrderedDict()
        for d in args:
            if isinstance(d, Dict):
                self._update(d)
            else:
                raise ValueError(f"{d} is not valid to initialize DataFrames")
        self._update(kwargs)
        self._col_to_df: Dict[str, Set[str]] = defaultdict(set)
        for k, v in self._data.items():
            for c in v:
                self._col_to_df[c].add(k)
        self._has_overlap = any(len(v) > 1 for v in self._col_to_df.values())

    @property
    def has_overlap(self):
        return self._has_overlap

    def get_col(self, expr) -> Tuple[str, str]:
        assert_or_throw(
            "*" not in expr, ValueError(f"{expr} is invalid to get a single column")
        )
        cols = self._get_cols(expr)
        assert_or_throw(len(cols) == 1, f"expected single column but get [{cols}]")
        return cols[0]

    def get_cols(
        self, *exprs: Any, ensure_distinct: bool = False, ensure_single_df: bool = False
    ) -> List[Tuple[str, str]]:
        res: List[Tuple[str, str]] = []
        for e in exprs:
            res += self._get_cols(e)
        if ensure_distinct:
            assert_or_throw(len(set(res)) == len(res), f"there are duplicates {res}")
        if ensure_single_df:
            assert_or_throw(
                len(set(x[0] for x in res)) == 1, f"not from single dataframe {res}"
            )
        return res

    def _update(self, dfs: Dict):
        for k, v in dfs.items():
            assert_or_throw(isinstance(k, str), ValueError(f"{k} is not string"))
            assert_or_throw(k != "", ValueError("key can't be empty"))
            if isinstance(v, DataFrame):
                self._data[k] = list(v.keys())
            elif isinstance(v, Iterable):
                self._data[k] = list(v)
            else:
                raise ValueError(f"{v} is invalid")

    def _get_cols(self, expr: str) -> List[Tuple[str, str]]:
        ee = expr.split(".", 1)
        if len(ee) == 1:
            df_name, col_name = "", unquote_name(expr)
        else:
            df_name, col_name = ee[0], unquote_name(ee[1])
        if col_name == "*":
            if df_name != "":
                return [(df_name, x) for x in self._data[df_name]]
            else:
                assert_or_throw(
                    not self.has_overlap, ValueError("There is schema overlap")
                )
                res: List[Tuple[str, str]] = []
                for k, v in self._data.items():
                    res += [(k, x) for x in v]
                return res
        elif df_name == "":
            dfs = self._col_to_df[col_name]
            assert_or_throw(len(dfs) > 0, ValueError(f"{col_name} is not defined"))
            assert_or_throw(
                len(dfs) == 1, ValueError(f"{col_name} in these dataframes {dfs}")
            )
            return [(next(iter(dfs)), col_name)]
        else:
            assert_or_throw(
                col_name in self._col_to_df and df_name in self._col_to_df[col_name],
                ValueError(f"{expr} does not exist"),
            )
            return [(df_name, col_name)]


class DataFrames(IndexedOrderedDict[str, DataFrame]):
    def __init__(self, *args: Any, **kwargs: Any):  # noqa: C901
        super().__init__()
        for d in args:
            if isinstance(d, Dict):
                self._update(d)
            else:
                raise ValueError(f"{d} is not valid to initialize DataFrames")
        self._update(kwargs)
        self.set_readonly()
        self._col_to_df: Dict[str, Set[str]] = defaultdict(set)
        for k, v in self.items():
            for c in v.keys():
                self._col_to_df[c].add(k)
        self._has_overlap = any(len(v) > 1 for v in self._col_to_df.values())
        self._resolver = ColumnNameResolver(self)

    @property
    def name_resolver(self) -> ColumnNameResolver:
        return self._resolver

    def get(self, expr: Union[str, List[str]]) -> Union[Column, "DataFrame"]:
        if isinstance(expr, str):
            return self._get_single_expr(expr)
        return DataFrame(*[self._get_single_expr(x) for x in expr])

    def get_columns(self, expr: Union[str, List[str]]) -> List[Column]:
        res = self.get(expr)
        if isinstance(res, Column):
            return [res]
        return list(res.values())

    def _get_single_expr(self, expr: str) -> Union[Column, "DataFrame"]:
        ee = expr.split(".", 1)
        if len(ee) == 1:
            df_name, col_name = "", unquote_name(expr)
        else:
            df_name, col_name = ee[0].strip(), unquote_name(ee[1])
        if col_name == "*":
            if df_name != "":
                return self[df_name][col_name]
            else:
                assert_or_throw(
                    not self._has_overlap, ValueError("There is schema overlap")
                )
                return DataFrame(*list(self.values()))
        elif df_name == "":
            dfs = self._col_to_df[col_name]
            assert_or_throw(len(dfs) > 0, ValueError(f"{col_name} is not defined"))
            assert_or_throw(
                len(dfs) == 1, ValueError(f"{col_name} in these dataframes {dfs}")
            )
            return self[next(iter(dfs))][col_name]
        else:
            df = self[df_name]
            assert_or_throw(
                col_name in df, ValueError(f"{df_name} does not have {col_name}")
            )
            return df[col_name]

    def _update(self, dfs: Dict):
        for k, v in dfs.items():
            assert_or_throw(isinstance(k, str), ValueError(f"{k} is not string"))
            assert_or_throw(k != "", ValueError("key can't be empty"))
            assert_or_throw(
                isinstance(v, DataFrame), ValueError(f"{v} is not DataFrame")
            )
            self[k] = v



# TODO: This is based on a duplication of Triad PandasLikeUtils
# should we remove PandasLikeUtils from Triad?
from typing import (
    Any,
    Generic,
    List,
    TypeVar,
    Tuple,
)

from triad import assert_or_throw


from triad.utils.pandas_like import PandasLikeUtils as PandasLikeUtilsBase

DfT = TypeVar("DfT", bound=Any)
ColT = TypeVar("ColT", bound=Any)

_ANTI_INDICATOR = "__anti_indicator__"
_CROSS_INDICATOR = "__corss_indicator__"


class PandasLikeUtils(PandasLikeUtilsBase[DfT], Generic[DfT, ColT]):
    """A collection of utils for sql operations on general pandas like dataframes"""

    def parse_join_type(self, join_type: str) -> str:
        join_type = join_type.replace(" ", "").replace("_", "").lower()
        if join_type in ["inner", "cross"]:
            return join_type
        if join_type in ["inner", "join"]:
            return "inner"
        if join_type in ["semi", "leftsemi"]:
            return "left_semi"
        if join_type in ["anti", "leftanti"]:
            return "left_anti"
        if join_type in ["left", "leftouter"]:
            return "left_outer"
        if join_type in ["right", "rightouter"]:
            return "right_outer"
        if join_type in ["outer", "full", "fullouter"]:
            return "full_outer"
        raise NotImplementedError(join_type)

    def drop_duplicates(self, df: DfT) -> DfT:
        return df.drop_duplicates()

    def union(self, ndf1: DfT, ndf2: DfT, unique: bool) -> DfT:
        ndf1, ndf2 = self._preprocess_set_op(ndf1, ndf2)
        ndf = self.concat_dfs(ndf1, ndf2)
        if unique:
            ndf = self.drop_duplicates(ndf)
        return ndf

    def intersect(self, df1: DfT, df2: DfT, unique: bool) -> DfT:
        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
        ndf = ndf1.merge(self.drop_duplicates(ndf2))
        if unique:
            ndf = self.drop_duplicates(ndf)
        return ndf

    def except_df(
        self,
        df1: DfT,
        df2: DfT,
        unique: bool,
        anti_indicator_col: str = _ANTI_INDICATOR,
    ) -> DfT:
        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
        # TODO: lack of test to make sure original ndf2 is not changed
        ndf2 = self._with_indicator(ndf2, anti_indicator_col)
        ndf = ndf1.merge(ndf2, how="left", on=list(ndf1.columns))
        ndf = ndf[ndf[anti_indicator_col].isnull()].drop([anti_indicator_col], axis=1)
        if unique:
            ndf = self.drop_duplicates(ndf)
        return ndf

    def join(
        self,
        ndf1: DfT,
        ndf2: DfT,
        join_type: str,
        on: List[str],
        anti_indicator_col: str = _ANTI_INDICATOR,
        cross_indicator_col: str = _CROSS_INDICATOR,
    ) -> DfT:
        join_type = self.parse_join_type(join_type)
        if join_type == "inner":
            ndf1 = ndf1.dropna(subset=on)
            ndf2 = ndf2.dropna(subset=on)
            joined = ndf1.merge(ndf2, how=join_type, on=on)
        elif join_type == "left_semi":
            ndf1 = ndf1.dropna(subset=on)
            ndf2 = self.drop_duplicates(ndf2[on].dropna())
            joined = ndf1.merge(ndf2, how="inner", on=on)
        elif join_type == "left_anti":
            # TODO: lack of test to make sure original ndf2 is not changed
            ndf2 = self.drop_duplicates(ndf2[on].dropna())
            ndf2 = self._with_indicator(ndf2, anti_indicator_col)
            joined = ndf1.merge(ndf2, how="left", on=on)
            joined = joined[joined[anti_indicator_col].isnull()].drop(
                [anti_indicator_col], axis=1
            )
        elif join_type == "left_outer":
            ndf2 = ndf2.dropna(subset=on)
            joined = ndf1.merge(ndf2, how="left", on=on)
        elif join_type == "right_outer":
            ndf1 = ndf1.dropna(subset=on)
            joined = ndf1.merge(ndf2, how="right", on=on)
        elif join_type == "full_outer":
            add: List[str] = []
            for f in on:
                name = f + "_null"
                s1 = ndf1[f].isnull().astype(int)
                ndf1[name] = s1
                s2 = ndf2[f].isnull().astype(int) * 2
                ndf2[name] = s2
                add.append(name)
            joined = ndf1.merge(ndf2, how="outer", on=on + add).drop(add, axis=1)
        elif join_type == "cross":
            assert_or_throw(
                len(on) == 0, ValueError(f"cross join can't have join keys {on}")
            )
            ndf1 = self._with_indicator(ndf1, cross_indicator_col)
            ndf2 = self._with_indicator(ndf2, cross_indicator_col)
            joined = ndf1.merge(ndf2, how="inner", on=[cross_indicator_col]).drop(
                [cross_indicator_col], axis=1
            )
        else:  # pragma: no cover
            raise NotImplementedError(join_type)
        return joined

    def concat_dfs(self, *dfs: DfT) -> DfT:
        raise NotImplementedError

    def _preprocess_set_op(self, ndf1: DfT, ndf2: DfT) -> Tuple[DfT, DfT]:
        assert_or_throw(
            len(list(ndf1.columns)) == len(list(ndf2.columns)),
            ValueError("two dataframes have different number of columns"),
        )
        ndf2.columns = ndf1.columns  # this is SQL behavior
        return ndf1, ndf2

    def _with_indicator(self, df: DfT, name: str) -> DfT:
        return df.assign(**{name: 1})



from pytest import raises
from triad.utils.hash import to_uuid

from qpd.dataframe import Column, DataFrame, DataFrames, ColumnNameResolver


def test_column():
    obj = "xx"
    c = Column(obj)
    assert obj is c.native
    assert "" == c.name
    # assert not c.has_name
    # raises(AssertionError, lambda: c.assert_has_name())
    c2 = c.rename("x")
    assert obj is c2.native
    assert "x" == c2.name
    assert to_uuid(c) != to_uuid(c2)
    c2.assert_has_name()
    c3 = Column(obj, "x")
    assert to_uuid(c3) == to_uuid(c2)


def test_dataframe():
    c1 = Column(1, "a")
    c2 = Column(1, "b")
    d = DataFrame(c1, c2)
    assert ["a", "b"] == list(d.keys())
    c3 = Column(1, "c")
    d2 = DataFrame(d, c3)
    assert ["a", "b", "c"] == list(d2.keys())
    d3 = DataFrame(c1, c2, c3)
    d4 = DataFrame(c1, [c2, c3])
    assert to_uuid(d2) != to_uuid(d)
    assert to_uuid(d2) == to_uuid(d3)
    assert to_uuid(d2) == to_uuid(d4)
    raises(ValueError, lambda: DataFrame(1))


def test_dataframe_get():
    c1 = Column(1, "a")
    c2 = Column(1, "b")
    d = DataFrame(c1, c2)
    assert d["a"] is c1
    assert d["*"] is d
    d2 = d[["b", "a"]]
    assert isinstance(d2, DataFrame)
    assert ["b", "a"] == list(d2.keys())
    assert to_uuid(d) != to_uuid(d2)
    d3 = d[["*"]]
    assert to_uuid(d3) == to_uuid(d)
    raises(ValueError, lambda: DataFrame(c1, c1))
    with raises(KeyError):
        d["x"]
    with raises(KeyError):
        d[["a", "x"]]
    with raises(ValueError):
        d[["a", "*"]]


def test_dataframes():
    c1 = Column(1, "a")
    c2 = Column(1, "b")
    c3 = Column(1, "a")
    c4 = Column(1, "c")
    df1 = DataFrame(c1, c2)
    df2 = DataFrame(c3, c4)
    df3 = DataFrame(c4)
    dfs1 = DataFrames(a=df1, b=df3)
    assert dfs1["a"] is df1
    assert dfs1["b"] is df3
    dfs2 = DataFrames(dict(a=df1), b=df3)
    assert to_uuid(dfs1) == to_uuid(dfs2)
    with raises(ValueError):
        DataFrames(1)
    dfs3 = DataFrames(dfs2, c=df2)
    assert dfs3["a"] is df1
    assert dfs3["b"] is df3
    assert dfs3["c"] is df2
    assert to_uuid(dfs1) != to_uuid(dfs3)


def test_dataframes_get():
    def assert_eq(df1, df2):
        assert list(df1.keys()) == list(df2.keys())
        assert to_uuid(df1) == to_uuid(df2)

    c1 = Column(1, "a")
    c2 = Column(1, "b")
    c3 = Column(1, "a")
    c4 = Column(1, "c")
    df1 = DataFrame(c1, c2)
    df2 = DataFrame(c3, c4)
    df3 = DataFrame(c4)
    dfs1 = DataFrames(a=df1, b=df3)
    assert_eq(dfs1.get("*"), DataFrame(c1, c2, c4))
    assert_eq(dfs1.get(["*"]), DataFrame(c1, c2, c4))
    assert_eq(dfs1.get(["c", "b"]), DataFrame(c4, c2))
    assert_eq(dfs1.get("a.*"), DataFrame(c1, c2))
    assert dfs1.get("a") is c1
    assert dfs1.get("a.a") is c1
    dfs2 = DataFrames(a=df1, b=df2)
    raises(ValueError, lambda: dfs2.get("*"))
    raises(ValueError, lambda: dfs2.get(["*"]))
    assert_eq(dfs2.get("a.*"), DataFrame(c1, c2))
    assert_eq(dfs2.get("b.*"), DataFrame(c3, c4))


def test_column_name_resolver():
    def assert_eq(expected, actual):
        s = [x[0] + "." + x[1] for x in actual]
        assert expected == ",".join(s)

    c = ColumnNameResolver(dict(a=["x", "y"]), b=["x", "z"])
    assert c.has_overlap
    assert_eq("a.x,a.y", c.get_cols("a.*", ensure_single_df=True))
    assert_eq("b.x,b.z", c.get_cols("b.x", "z", ensure_single_df=True))
    assert_eq("b.z,a.x,a.y", c.get_cols("z", "a.x", "y"))
    assert_eq("b.z,a.x,b.z", c.get_cols("z", "a.x", "z"))
    raises(AssertionError, lambda: c.get_cols("z", "a.x", "z", ensure_distinct=True))
    raises(AssertionError, lambda: c.get_cols("z", "a.x", ensure_single_df=True))
    raises(ValueError, lambda: c.get_cols("t"))
    raises(ValueError, lambda: c.get_cols("*"))

    assert ("a", "y") == c.get_col("y")
    raises(ValueError, lambda: c.get_col("x"))
    raises(ValueError, lambda: c.get_col("*"))
    raises(ValueError, lambda: c.get_col("a.*"))

    raises(ValueError, lambda: ColumnNameResolver(1))
    raises(ValueError, lambda: ColumnNameResolver(a=1))

    c1 = Column(1, "x")
    c2 = Column(1, "y")
    df = DataFrame(c1, c2)
    c = ColumnNameResolver(dict(b=df), a=["m", "n"])
    assert_eq("b.x,b.y,a.m,a.n", c.get_cols("*"))

    c = ColumnNameResolver(dict(a=["-", "y"]), b=["-", "z"])
    assert c.has_overlap
    assert_eq("a.-,a.y", c.get_cols("a.*", ensure_single_df=True))
    assert_eq("a.-,a.y", c.get_cols("a.`-`", "y", ensure_single_df=True))



from typing import Set

JOIN_TYPES: Set[str] = {
    "inner",
    "cross",
    "left_semi",
    "left_anti",
    "left_outer",
    "right_outer",
    "full_outer",
}

AGGREGATION_FUNCTIONS: Set[str] = {
    "count",
    "mean",
    "avg",
    "min",
    "max",
    "sum",
    "first",
    "first_value",
    "last",
    "last_value",
}

WINDOW_FUNCTIONS: Set[str] = AGGREGATION_FUNCTIONS.union(
    {
        "rank",
        "dense_rank",
        "percentile_rank",
        "percent_rank",
        "row_number",
        "lag",
        "lead",
    }
)



from typing import Any

import pandas as pd

from qpd_pandas import QPDPandasEngine
from qpd_test.benchmark_suite import BenchmarkTests


class PandasBenchmarkTests(BenchmarkTests.Tests):
    def _test_select(self):
        for n, q, s in self._test_select(10, 100, 1000, 10000, repeat=3):
            print(n, q, s)

    def test_select_where(self):
        for n, q, s in self._test_select_where(10, 100, 1000, 10000, 100000, repeat=3):
            print(n, q, s)

    def _test_integration_1(self):
        for n, q, s in self._test_integration_1(10, 100, 1000, repeat=3):
            print(n, q, s)

    def make_qpd_engine(self):
        return QPDPandasEngine()

    def to_native_df(self, data: Any, columns: Any) -> Any:  # pragma: no cover
        if isinstance(data, pd.DataFrame):
            return data
        return pd.DataFrame(data, columns=columns)



from typing import Any

import pandas as pd
from qpd.dataframe import Column, DataFrame
from qpd_pandas import QPDPandasEngine
from qpd_pandas import run_sql_on_pandas
from qpd_test.sql_suite import SQLTests
from qpd_test.utils import assert_df_eq


def test_entrypoint():
    pdf = pd.DataFrame([[0, 1], [1, 2], [0, 4]], columns=["a", "b"])
    df1 = run_sql_on_pandas(
        "SELECT a, SUM(b) AS b, COUNT(*) AS c FROM df GROUP BY a", {"df": pdf}
    )
    df2 = run_sql_on_pandas("select a+1 as a, b from df", {"df": pdf}, ignore_case=True)
    res = run_sql_on_pandas(
        """
        SELECT a.a, a.b, c, b.b AS d 
        FROM a INNER JOIN b ON a.a=b.a
        """,
        dict(a=df1, b=df2),
    )
    assert_df_eq(res, [[1, 2, 1, 1], [1, 2, 1, 4]], list("abcd"))


class SQLTests(SQLTests.Tests):
    def to_pandas_df(self, df: Any) -> pd.DataFrame:
        if isinstance(df, DataFrame):
            return self.op.to_native(df)
        elif isinstance(df, pd.DataFrame):
            return df
        else:
            raise ValueError(df)

    def to_pandas_series(self, col: Column) -> pd.Series:
        return col.native

    def make_qpd_engine(self):
        return QPDPandasEngine()

    def to_native_df(self, data: Any, columns: Any) -> Any:  # pragma: no cover
        if isinstance(data, pd.DataFrame):
            return data
        return pd.DataFrame(data, columns=columns)
